"""
/***************************************************************************
 InspireAtomClientDialog
                                 A QGIS plugin
 Client for INSPIRE Downloadservices based on ATOM-Feeds
                             -------------------
        begin                : 2012-05-28
        copyright            : (C) 2012 by Juergen Weichand
        email                : juergen@weichand.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from PyQt4 import QtCore, QtGui
from PyQt4.QtNetwork import QHttp
from ui_inspireatomclient import Ui_InspireAtomClient
from qgis.core import *
from xml.etree import ElementTree
from urlparse import urljoin
from urlparse import urlparse
import urllib2 
import string
import random
import tempfile
import os
import os.path
import inspireatomlib


# create the dialog for zoom to point
class InspireAtomClientDialog(QtGui.QDialog):
    
    def __init__(self, parent):
        QtGui.QDialog.__init__(self)
        # Set up the user interface from Designer.
        self.parent = parent
        self.ui = Ui_InspireAtomClient()
        self.ui.setupUi(self)

        self.settings = QtCore.QSettings()
        self.init_variables()

        QtCore.QObject.connect(self.ui.cmdGetFeed, QtCore.SIGNAL("clicked()"), self.get_service_feed)
        QtCore.QObject.connect(self.ui.cmdSelectDataset, QtCore.SIGNAL("clicked()"), self.select_dataset_feed)
        QtCore.QObject.connect(self.ui.cmdDownload, QtCore.SIGNAL("clicked()"), self.download_files)
        QtCore.QObject.connect(self.ui.cmbDatasetRepresentations, QtCore.SIGNAL("currentIndexChanged(int)"), self.update_lw_files)

    def init_variables(self):
        self.onlineresource = ""
        self.datasetrepresentations = {}
        self.currentfile = 0
        self.downloadedfiles = []

    # request and handle "Service Feed" - Get Metadata | cmdGetFeed Signal
    def get_service_feed(self):
        self.ui.cmdGetFeed.setEnabled(False)
        self.init_variables()
        try:
            self.setup_proxy()
            self.onlineresource = self.ui.txtUrl.text().trimmed()
            request = unicode(self.onlineresource)
            response = urllib2.urlopen(request)
            buf = response.read()
        except urllib2.HTTPError, e:  
            self.ui.lblMessage.setText("HTTP Error: {0}".format(e.code))   
            self.ui.cmdGetFeed.setEnabled(True)
        except urllib2.URLError, e:
            self.ui.lblMessage.setText("URL Error: {0}".format(e.reason))
            self.ui.cmdGetFeed.setEnabled(True)
        else:
            layername="INSPIRE_DLS#{0}".format(''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(6)))
            tmpfile = self.save_tempfile("{0}.rss".format(layername), buf)
            vlayer = QgsVectorLayer(tmpfile, layername, "ogr")
            if not vlayer.isValid():
                self.ui.lblMessage.setText("Response is not a valid QGIS-Layer!")
                self.ui.cmdGetFeed.setEnabled(True)
            else: 
                QgsMapLayerRegistry.instance().addMapLayer(vlayer)
                self.parent.iface.zoomToActiveLayer()
                self.ui.cmdGetFeed.setEnabled(False)
                self.ui.cmdSelectDataset.setEnabled(True)


    # select "Dataset Feed" | cmdSelectDataset Signal
    def select_dataset_feed(self):
        self.clear_frame()
        self.ui.lblMessage.setText("")
        # http://www.qgisworkshop.org/html/workshop/plugins_tutorial.html
        result = QtCore.QObject.connect(self.parent.clickTool, QtCore.SIGNAL("canvasClicked(const QgsPoint &, Qt::MouseButton)"), self.select_dataset_feed_procedure)
        self.parent.iface.mapCanvas().setMapTool(self.parent.clickTool)


    # select "Dataset Feed" | Signal ("Click")
    def select_dataset_feed_procedure(self, point, button):
        self.clear_frame()
        # setup the provider select to filter results based on a rectangle
        pntGeom = QgsGeometry.fromPoint(point)  
        # scale-dependent buffer of 2 pixels-worth of map units
        pntBuff = pntGeom.buffer( (self.parent.iface.mapCanvas().mapUnitsPerPixel() * 2), 0) 
        rect = pntBuff.boundingBox()
        # get currentLayer and dataProvider
        cLayer = self.parent.iface.mapCanvas().currentLayer()
        if cLayer.type() != QgsMapLayer.VectorLayer:
            return 
        selectList = []
        provider = cLayer.dataProvider()
        feat = QgsFeature()
        # create the select statement
        provider.select(provider.attributeIndexes(), rect)
        while provider.nextFeature(feat):
            # if the feat geom returned from the selection intersects our point then put it in a list
            if feat.geometry().intersects(pntGeom):
                selectList.append(feat.id())
                map = feat.attributeMap()               
                try:
                    if len(unicode(map[provider.fieldNameIndex("inspire_dls_spatial_dataset_identifier_code")].toString())) > 0:
                        dataset = inspireatomlib.Dataset(unicode(map[provider.fieldNameIndex("inspire_dls_spatial_dataset_identifier_code")].toString()))
                        if provider.fieldNameIndex("title") > -1:
                            dataset.setTitle(unicode(map[provider.fieldNameIndex("title")].toString()))
                        if provider.fieldNameIndex("summary") > -1:
                            dataset.setSummary(unicode(map[provider.fieldNameIndex("summary")].toString()))                    
                        if provider.fieldNameIndex("rights") > -1:
                            dataset.setRights(unicode(map[provider.fieldNameIndex("rights")].toString()))
                
                        for key, value in map.items():                        
                            if value.toString() == "alternate":
                                fieldname = provider.fields()[key].name().replace("rel", "href")
                                if provider.fieldNameIndex(fieldname) > -1:
                                    linksubfeed = unicode(map[provider.fieldNameIndex(fieldname)].toString())
                                    dataset.setLinkSubfeed(self.buildurl(linksubfeed))
                            if value.toString() == "describedby":
                                fieldname = provider.fields()[key].name().replace("rel", "href")
                                if provider.fieldNameIndex(fieldname) > -1:
                                    linkmetadata = unicode(map[provider.fieldNameIndex(fieldname)].toString())
                                    dataset.setLinkMetadata(self.buildurl(linkmetadata))
                
                        self.ui.cmbDatasetRepresentations.clear()
                        self.ui.frmDataset.setEnabled(True)
                        self.ui.lblTitle.setText(dataset.getTitle())
                        self.ui.txtSummary.setPlainText(dataset.getSummary())
                        self.ui.txtId.setPlainText(dataset.getId())
                        self.ui.txtRights.setPlainText(dataset.getRights())                    
                        self.receive_dataset_representations(dataset.getLinkSubfeed())
                except KeyError, e:
                    self.ui.lblMessage.setText("Invalid INSPIRE Service Feed Feature selected!")
                    
        # make the actual selection
        cLayer.setSelectedFeatures(selectList)
        result = QtCore.QObject.disconnect(self.parent.clickTool, QtCore.SIGNAL("canvasClicked(const QgsPoint &, Qt::MouseButton)"), self.select_dataset_feed_procedure)
        self.parent.iface.mapCanvas().unsetMapTool(self.parent.clickTool)

    # request and handle "Dataset Feed" (dataset representations)
    def receive_dataset_representations(self, subfeedurl): 
        try:
            self.setup_proxy()
            response = urllib2.urlopen(subfeedurl)
            buf = response.read()
        except urllib2.HTTPError, e:  
            self.ui.lblMessage.setText("HTTP Error: {0}".format(e.code))   
        except urllib2.URLError, e:
            self.ui.lblMessage.setText("URL Error: {0}".format(e.reason))
        else:
            self.datasetrepresentations = {}
            root = ElementTree.fromstring(buf)
            # ATOM Namespace
            namespace = "{http://www.w3.org/2005/Atom}"
            # check correct Rootelement 
            if root.tag == "{0}feed".format(namespace):       
                for target in root.findall("{0}entry".format(namespace)):
                    idvalue = ""
                    for id in target.findall("{0}id".format(namespace)):
                        idvalue = id.text
                    if (idvalue):
                        datasetrepresentation = inspireatomlib.DatasetRepresentation(idvalue)
                        for title in target.findall("{0}title".format(namespace)):
                            datasetrepresentation.setTitle(title.text)
                        files = []
                        for link in target.findall("{0}link".format(namespace)):
                            if link.get("rel") == "alternate":
                                files.append(self.buildurl(link.get("href")))                           
                            if link.get("rel") == "section":
                                files.append(self.buildurl(link.get("href")))
                        datasetrepresentation.setFiles(files)
                        self.datasetrepresentations[QtCore.QString(datasetrepresentation.getTitle())] = datasetrepresentation
                        self.ui.cmbDatasetRepresentations.addItem(datasetrepresentation.getTitle(), datasetrepresentation.getTitle())
                        self.ui.cmdDownload.setEnabled(True)
            
            else:
                self.ui.lblMessage.setText("Invalid INSPIRE Dataset Feed returned!")  

    # update ListWidget | cmbDatasetRepresentations "currentIndexChanged(int)" Signal
    def update_lw_files(self):
        self.ui.lwFiles.clear()
        if len(self.ui.cmbDatasetRepresentations.currentText()) > 0:
            datasetrepresentation = self.datasetrepresentations[self.ui.cmbDatasetRepresentations.currentText()]
            for file in datasetrepresentation.getFiles():                
                self.ui.lwFiles.addItem(file)

    def clear_frame(self):
        self.ui.lblTitle.setText("Title")
        self.ui.txtSummary.setPlainText("")
        self.ui.txtId.setPlainText("")
        self.ui.txtRights.setPlainText("")
        self.ui.cmbDatasetRepresentations.clear()
        self.ui.lwFiles.clear()
        self.ui.cmdDownload.setEnabled(False)
        self.ui.frmDataset.setEnabled(False)

    """
    ############################################################################################################################
    # Download
    ############################################################################################################################
    """
   
    # execute download procedure | cmdDownload Signal
    def download_files(self):
        self.ui.lblMessage.setText("")

        self.httpGetId = 0
        self.httpRequestAborted = False
        self.downloadedfiles = []

        self.setup_qhttp()
        self.http.requestFinished.connect(self.httpRequestFinished)
        self.http.dataReadProgress.connect(self.updateDataReadProgress)
        self.http.responseHeaderReceived.connect(self.readResponseHeader)

        self.download_next()

    # download next file (after finishing the last one)
    def download_next(self):
       datasetrepresentation = self.datasetrepresentations[self.ui.cmbDatasetRepresentations.currentText()]
       if self.currentfile < len(datasetrepresentation.getFiles()):
           self.ui.cmdDownload.setEnabled(False)
           self.ui.cmdSelectDataset.setEnabled(False)
           self.ui.lwFiles.setItemSelected(self.ui.lwFiles.item(self.currentfile), True)
           self.ui.cmdDownload.setText("Downloading {0}/{1}".format(self.currentfile + 1, len(datasetrepresentation.getFiles())))
           file = datasetrepresentation.getFiles()[self.currentfile]
           filename=self.buildfilename(file)
           self.downloadFile(file, self.get_temppath(filename))
           self.currentfile += 1
       else:
           self.load_downloaded_files()
           self.reset_ui_download()

    # try to load downloaded files as QGIS-Layer(s)
    def load_downloaded_files(self):
        failed = []
        successful = []
        for file in self.downloadedfiles:
            vlayer = QgsVectorLayer(file, file, "ogr")            
            if not vlayer.isValid():
                #fileInfo = QtCore.QFileInfo(file)
                #baseName = fileInfo.baseName()
                rlayer = QgsRasterLayer(file, file)   
                if not rlayer.isValid():
                    failed.append(file)
                    self.ui.lblMessage.setText("")
                else:
                    QgsMapLayerRegistry.instance().addMapLayer(rlayer)
                    self.parent.iface.zoomToActiveLayer()
                    successful.append(file)
            else: 
                self.ui.lblMessage.setText("")
                QgsMapLayerRegistry.instance().addMapLayer(vlayer)
                self.parent.iface.zoomToActiveLayer()
                successful.append(file)

        
        message = ""
        if len(successful) > 0:
            message += "<p><b>Successfully loaded:</b><br />"
            for file in successful:
                message += file + "<br />"
            message += "</p>"        
        if len(failed) > 0:
            message += "<p><b>Failed to load:</b><br />"
            for file in failed:
                message += file + "<br />"
            message += "</p>"
        QtGui.QMessageBox.information(self, "Import Status", message)

    """
    ############################################################################################################################
    # UTIL
    ############################################################################################################################
    """

    def get_temppath(self, filename):
        tmpdir = os.path.join(tempfile.gettempdir(),'inspireatomclient')
        if not os.path.exists(tmpdir):
            os.makedirs(tmpdir)
        tmpfile= os.path.join(tmpdir, filename)
        return tmpfile


    def save_tempfile(self, filename, content):
        tmpdir = os.path.join(tempfile.gettempdir(),'inspireatomclient')
        if not os.path.exists(tmpdir):
            os.makedirs(tmpdir)
        tmpfile= os.path.join(tmpdir, filename)
        fobj=open(tmpfile,'wb')
        fobj.write(content)
        fobj.close()  
        return tmpfile

    # Receive Proxy from QGIS-Settings
    def getProxy(self):
        if self.settings.value("/proxy/proxyEnabled").toString() == "true":
           proxy = "{0}:{1}".format(self.settings.value("/proxy/proxyHost").toString(), self.settings.value("/proxy/proxyPort").toString())
           if proxy.startswith("http://"):
               return proxy
           else:
               return "http://" + proxy
        else: 
            return ""
    
    # Setup urllib2 (Proxy)
    def setup_proxy(self):
        if not self.getProxy() == "":
            proxy_support = urllib2.ProxyHandler({"http" : self.getProxy()})
            opener = urllib2.build_opener(proxy_support)
            urllib2.install_opener(opener)
        else: 
            proxy_support = urllib2.ProxyHandler({})
            opener = urllib2.build_opener(proxy_support)
            urllib2.install_opener(opener)


    # Setup Qhttp (Proxy)
    def setup_qhttp(self):
        self.http = QHttp(self)
        if not self.getProxy() == "":
            self.http.setProxy(QgsNetworkAccessManager.instance().fallbackProxy()) # Proxy       

    # Convert relative links to absolute link
    def buildurl(self, urlfragment):
        if not urlfragment.startswith("http"):
            return urljoin(unicode(self.onlineresource), urlfragment)
        return urlfragment

    # Build filename for downloaded file
    def buildfilename(self, url):
        parseresult = urlparse(url)
        if len(parseresult.query) == 0:
            path = parseresult.path
            filename = path[path.rfind("/") + 1:]
        else:
            extension="ext"
            if url.lower().find("wfs") > -1:
                extension="gml"
            if url.lower().find("wms") > -1:
                extension="tiff"
            filename="{0}.{1}".format(''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(6)), extension)
        return filename


    #############################################################################################################
    # QHttp GetFeature-Request - http://stackoverflow.com/questions/6852038/threading-in-pyqt4
    #############################################################################################################

    def downloadFile(self, httplink, fileName):
        #self.lock_ui()
        url = QtCore.QUrl(httplink)

        if QtCore.QFile.exists(fileName):
            QtCore.QFile.remove(fileName)

        self.outFile = QtCore.QFile(fileName)
        if not self.outFile.open(QtCore .QIODevice.WriteOnly):
            QtGui.QMessageBox.information(self, 'Error',
                    'Unable to save the file %s: %s.' % (fileName, self.outFile.errorString()))
            self.outFile = None
            return

        mode = QHttp.ConnectionModeHttp
        port = url.port()
        if port == -1:
            port = 0
        self.http.setHost(url.host(), mode, port)
        self.httpRequestAborted = False
        # Download the file.
        self.ui.progressBar.setVisible(True)
        self.httpGetId = self.http.get(url.toString(), self.outFile)

    # currently unused
    def cancelDownload(self):
        self.httpRequestAborted = True
        self.http.abort()
        self.close()

        self.ui.progressBar.setMaximum(1)
        self.ui.progressBar.setValue(0)
        self.unlock_ui()

    def httpRequestFinished(self, requestId, error):
        if requestId != self.httpGetId:
            return

        if self.httpRequestAborted:
            if self.outFile is not None:
                self.outFile.close()
                self.outFile.remove()
                self.outFile = None
            return

        self.outFile.close()

        self.ui.progressBar.setMaximum(1)
        self.ui.progressBar.setValue(1)

        if error:
            self.outFile.remove()
            self.ui.lblMessage.setText("Download failed: %s." % self.http.errorString())
            self.reset_ui_download()
        else:      
            self.ui.lblMessage.setText("")
            self.downloadedfiles.append(unicode(self.outFile.fileName()))
            self.download_next()

        self.ui.progressBar.setMaximum(1)
        self.ui.progressBar.setValue(0)

    def readResponseHeader(self, responseHeader):
        # Check for genuine error conditions.
        if responseHeader.statusCode() not in (200, 300, 301, 302, 303, 307):
            QMessageBox.information(self, 'Error',
                    'Download failed: %s.' % responseHeader.reasonPhrase())
            self.httpRequestAborted = True
            self.http.abort()

    def updateDataReadProgress(self, bytesRead, totalBytes):
        if self.httpRequestAborted:
            return
        self.ui.progressBar.setMaximum(totalBytes)
        self.ui.progressBar.setValue(bytesRead)
        self.ui.lblMessage.setText("Please wait while downloading - {0} Bytes downloaded!".format(str(bytesRead)))

    def reset_ui_download(self):
        self.ui.cmdDownload.setText("Download")
        self.ui.cmdDownload.setEnabled(True)
        self.ui.cmdSelectDataset.setEnabled(True)
        self.currentfile = 0
    